# core/config.py

"""
Handle Config init, dependencies check, logging setup,
files and folders structures, config settings parameters init
"""

from pathlib import Path
import logging
import os
import tomllib
import subprocess
import socket

from rcdl.core.file_io import write_txt


class Config:
    """Global app var/parameters"""

    # paths
    APP_NAME = "rcdl"

    BASE_DIR = Path(os.environ.get("RCDL_BASE_DIR", Path.home() / "Videos/rcdl"))

    CACHE_DIR = BASE_DIR / ".cache"
    DB_PATH = CACHE_DIR / "cdl.db"
    LOG_FILE = CACHE_DIR / "cdl.log"
    CREATORS_FILE = CACHE_DIR / "creators.txt"
    DISCOVER_DIR = CACHE_DIR / "discover"
    CONFIG_FILE = CACHE_DIR / "config.toml"

    DEBUG = False
    DRY_RUN = False

    # api settings
    POST_PER_PAGE: int = 50
    DEFAULT_MAX_PAGE: int = 10
    MAX_FAIL_COUNT: int = 7
    TIMEOUT: int = 10

    # fuse settings
    MAX_WIDTH: int = 1920
    MAX_HEIGHT: int = 1080
    FPS: int = 30
    PRESET: str = "veryfast"
    THREADS: int = 0

    HANDBRAKE_RUN_CMD = "HandBrakeCLI"

    CHECKSUM_RETRY = 2

    HOST = "127.0.0.1"

    @classmethod
    def ensure_dirs(cls):
        """Ensure directory exist"""
        cls.CACHE_DIR.mkdir(parents=True, exist_ok=True)
        cls.DISCOVER_DIR.mkdir(exist_ok=True)

    @classmethod
    def ensure_files(cls):
        """Ensure file exist, populate default if necessary"""
        files = [cls.DB_PATH, cls.CREATORS_FILE, cls.CONFIG_FILE]
        for file in files:
            if not file.exists():
                file.touch()
                logging.info("Created file %s", file)
                if file == cls.CREATORS_FILE:
                    write_txt(cls.CREATORS_FILE, DEFAULT_CREATORS, mode="w")
                if file == cls.CONFIG_FILE:
                    write_txt(cls.CONFIG_FILE, DEFAULT_CONFIG, mode="w")

    @classmethod
    def creator_folder(cls, creator_id: str) -> Path:
        """Return creator folder path base on user/creator_id"""
        folder = cls.BASE_DIR / creator_id
        folder.mkdir(exist_ok=True)
        return folder

    @classmethod
    def cache_file(cls, filename: str, ext: str = ".json") -> Path:
        """Return filepath of a file in the .cache/ folder"""
        file_name = filename + ext
        file = cls.CACHE_DIR / file_name
        return file

    @classmethod
    def set_debug(cls, debug: bool):
        """Set class variable DEBUG"""
        cls.DEBUG = debug

    @classmethod
    def set_dry_run(cls, dry_run: bool):
        """Set class variable DRY_RUN"""
        cls.DRY_RUN = dry_run

    @classmethod
    def set_host(cls):
        cls.HOST = cls.get_host_ip()
        print("HOST:", cls.HOST)

    @staticmethod
    def get_host_ip():
        """
        Returns the LAN IP of the current machine.
        Falls back to 127.0.0.1 if unable to detect.
        """
        try:
            # connect to a dummy address to get the outgoing IP
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(("8.8.8.8", 80))  # Google DNS
            ip = s.getsockname()[0]
            s.close()
            return ip
        except Exception:
            return "127.0.0.1"

    @classmethod
    def load_config(cls):
        """Load config.toml and set class var with value in config.toml"""
        with open(cls.CONFIG_FILE, "rb") as f:
            data = tomllib.load(f)
        app = data.get("app", {})
        cls.DEFAULT_MAX_PAGE = app.get("default_max_page", cls.DEFAULT_MAX_PAGE)
        cls.MAX_FAIL_COUNT = app.get("max_fail_count", cls.MAX_FAIL_COUNT)
        cls.TIMEOUT = app.get("timeout", cls.TIMEOUT)
        cls.CHECKSUM_RETRY = app.get("checksum_retry", cls.CHECKSUM_RETRY)

        video = data.get("video", {})
        cls.MAX_WIDTH = video.get("max_width", cls.MAX_WIDTH)
        cls.MAX_HEIGHT = video.get("max_height", cls.MAX_HEIGHT)
        cls.FPS = video.get("fps", cls.FPS)
        cls.PRESET = video.get("preset", cls.PRESET)
        cls.THREADS = video.get("threads", cls.THREADS)

        paths = data.get("paths", {})
        if "base_dir" in paths:
            cls.BASE_DIR = Path(
                os.environ.get("RCDL_BASE_DIR", os.path.expanduser(paths["base_dir"]))
            )
            cls.CACHE_DIR = cls.BASE_DIR / ".cache"
        if "handbrake_run_cmd" in paths:
            cls.HANDBRAKE_RUN_CMD = paths.get("handbrake_run_cmd")


def setup_logging(log_file: Path, level: int = 0):
    """Setup logging for rcdl"""
    logger = logging.getLogger()
    logger.setLevel(level)
    logger.handlers.clear()  # avoid double handlers if called multiple times

    # loggin format & file handler
    file_handler = logging.FileHandler(log_file, encoding="utf-8", mode="a")
    file_handler.setFormatter(
        logging.Formatter(
            "{asctime} - {levelname} - {message}",
            style="{",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
    )
    logger.addHandler(file_handler)

    # log library warning/errors
    stream = logging.StreamHandler()
    stream.setLevel(logging.ERROR)  # only show warnings/errors from libraries
    logger.addHandler(stream)


def check_dependencies():
    """Check external program version against last tested working version"""
    for prgrm, info in DEPENDENCIES_TEST_VERSION.items():
        try:
            result = subprocess.run(
                info["cmd"],
                capture_output=True,
                text=True,
                shell=True,
                check=False,
            )
            version = result.stdout.strip()

            if version != info["version"]:
                print(
                    f"Last tested version for {prgrm}:"
                    f" {info['version']} -> yours: {version}"
                )
                if version == "":
                    print(f"{prgrm} is not installed.")
                print(f"Check {prgrm} is installed if your version is empty.")
        except (OSError, subprocess.SubprocessError) as e:
            print(
                f"Failed to check {prgrm} version due to: {e}\nCheck {prgrm} is installed."
            )


DEPENDENCIES_TEST_VERSION = {
    "yt-dlp": {"cmd": "yt-dlp --version", "version": "2025.12.08"},
    "aria2c": {
        "cmd": "aria2c -v | head -n 1",
        "version": "aria2 version 1.37.0",
    },
    "ffmpeg": {
        "cmd": 'ffmpeg -version | sed -n "s/ffmpeg version \\([-0-9.]*\\).*/\\1/p;"',
        "version": "7.1.1-1",
    },
    "handbrake": {
        "cmd": Config.HANDBRAKE_RUN_CMD
        + ' --version 2>&1 | sed -n "s/HandBrake \\([0-9.]*\\).*/\\1/p"',
        "version": "1.9.2",
    },
}


# default creators
DEFAULT_CREATORS = ["boixd/onlyfans"]

# default config params
DEFAULT_CONFIG: str = """\
[app]
default_max_page = 10
max_fail_count = 7
timeout = 10
checksum_retry = 2

[fuse]
max_width = 1920
max_height = 1080
fps = 30
preset = "veryfast"
threads = 0

[paths]
base_dir = "~/Videos/rcdl"
handbrake_run_cmd = "HandBrakeCLI"
"""



# gui/video_manager.py

import os
import urllib.parse

import streamlit as st
import streamlit.components.v1 as components


from rcdl.core.config import Config
from rcdl.core.models import Status, Media
from rcdl.core.db import DB
from rcdl.utils import format_seconds


previous_statuses = {}


def video_html(path):
    safe_path = urllib.parse.quote(path, safe="")
    # Use the same IP as the one in the browser URL

    return f"""
    <video controls preload="metadata" style="width:100%; max-height:320px;">
        <source
            src="http://{Config.HOST}:8123/video?path={safe_path}"
            type="video/mp4"
        >
    </video>
    """


def set_status(media: Media, status: Status):
    key = media.post_id + media.url
    previous_statuses[key] = media.status
    media.status = status

    with DB() as db:
        db.update_media(media)
    print(f"Set {media.post_id} to {status.value}")

    for m in st.session_state.medias:
        if m.post_id == media.post_id and m.url == media.url:
            m.status = status
            break


def video_manager():
    if "reload_medias" not in st.session_state:
        st.session_state.reload_medias = True
    if "media_index" not in st.session_state:
        st.session_state.media_index = 0

    st.title("Video Manager")
    Config.set_host()

    # Filter & Sorting UI
    with st.expander("Filters & Sorting", expanded=True):
        col1, col2, col3 = st.columns(3)
        with col1:
            st.selectbox(
                "Sort By",
                options=["file_size", "service", "duration", "file_path"],
                index=0,
                key="sort_by",
            )
        with col2:
            st.radio(
                "Order",
                options=[True, False],
                format_func=lambda x: "Ascending" if x else "Descending",
                horizontal=True,
                key="ascending",
            )
        with col3:
            st.text_input(
                "Creator ID(user)",
                placeholder="Leave empty for all",
                key="creator_filter",
            )
        st.multiselect(
            "Status",
            options=list(Status),
            default=[Status.DOWNLOADED, Status.OPTIMIZED],
            key="status_filter",
        )

        if st.button("Apply"):
            st.session_state.reload_medias = True

    # load db
    if st.session_state.reload_medias or "medias" not in st.session_state:
        with DB() as db:
            medias = db.query_medias_by_status_sorted(
                st.session_state.status_filter,
                sort_by=st.session_state.sort_by,
                ascending=st.session_state.ascending,
            )

            # check if in a fuse group

            # creator filter
            if st.session_state.creator_filter:
                filtered = []
                for m in medias:
                    post = db.query_post_by_id(m.post_id)
                    if (
                        post
                        and post.user == st.session_state.creator_filter
                        and db.query_fuses_by_id(m.post_id) is None
                    ):
                        filtered.append(m)
                medias = filtered

            st.session_state.medias = medias
            st.session_state.media_index = 0
        st.session_state.reload_medias = False

    medias = st.session_state.medias
    if not medias:
        st.info("No media found")
        return

    # session state
    if "media_index" not in st.session_state:
        st.session_state.media_index = 0

    idx = st.session_state.media_index
    media = medias[idx]

    # media info
    st.subheader(f"Media {idx + 1} / {len(medias)}")

    with DB() as db:
        post = db.query_post_by_id(media.post_id)
        if post is None:
            st.info("No matching post found")
            return

    col_video, col_info = st.columns([1, 2])
    with col_info:
        col1, col2 = st.columns(2)
        with col1:
            st.write("**Post ID:**", media.post_id)
            st.write("**Service:**", media.service)
            st.write("**User:**", post.user)
            st.write("**Duration:**", format_seconds(media.duration))
            st.write("**Sequence:**", media.sequence)
            st.write("**Size:**", round(media.file_size / (1024 * 1024), 1), "MB")
            st.write("**Status:**", media.status)
            key = media.post_id + media.url
            if key in previous_statuses:
                st.write("**PREV STATUS:**", previous_statuses[key])
            st.write("**Path:**", media.file_path)
            st.write("**Created at**:", media.created_at[0:16])

        with col2:
            # controls
            c1, c2, c3 = st.columns([1, 1, 2])
            with c1:
                if st.button("⏮ Prev", disabled=idx == 0):
                    st.session_state.media_index -= 1
                    st.rerun()
                if st.button("⏭ Next", disabled=idx >= len(medias) - 1):
                    st.session_state.media_index += 1
                    st.rerun()
            with c2:
                if st.button("Remove"):
                    set_status(media, Status.TO_BE_DELETED)
                    st.rerun()
                if st.button("Revert Status"):
                    key = media.post_id + media.url
                    if key in previous_statuses:
                        set_status(media, previous_statuses[key])
                    else:
                        print("Not in previous status")
                    st.rerun()
            with c3:
                chosen_status = st.selectbox(
                    "Set Status",
                    options=list(Status),
                    index=list(Status).index(media.status)
                    if media.status in list(Status)
                    else 0,
                )
                if st.button("Apply Status"):
                    set_status(media, chosen_status)
                    st.rerun()

    # video player
    with col_video:
        full_path = os.path.join(Config.creator_folder(post.user), media.file_path)

        if os.path.exists(full_path):
            with st.container(key=f"container_{media.post_id}_{idx}"):
                components.html(video_html(full_path), height=320)
        else:
            st.error(f"Video not found: {full_path}")



# core/adapters.py

"""Convert [Any] into proper Models from models.py"""

import json
import sqlite3
from dataclasses import fields

from rcdl.core import parser
from rcdl.interface.ui import UI
from rcdl.core.models import Post, Media, Status, FusedMedia
from rcdl.utils import get_date_now, get_json_hash

VALID_POST_KEYS = set(
    [
        "id",
        "user",
        "service",
        "title",
        "substring",
        "published",
        "file",
        "attachments",
    ]
)


def _compute_json_metadata(raw: dict) -> tuple[str, str, str]:
    """From a json dict, return:
    - raw_json: str
    - json_hash: str
    - fetched_at str (datetime)
    """
    raw_json, json_hash = get_json_hash(raw)
    fetched_at = get_date_now()
    return raw_json, json_hash, fetched_at


def json_posts_to_posts(posts: list[dict]) -> list[Post]:
    """Convert a list of json post (dict) into a list of Post model
    Ignore if conversion failed"""
    formatted_posts = []
    for post in posts:
        p = json_post_to_post(post)
        if p is not None:
            formatted_posts.append(p)
    return formatted_posts


def json_post_to_post(post: dict) -> Post | None:
    """Convert a json post (dict) into a Post model
    or return None if covnersion failed"""
    post_keys = set(post)
    if post_keys != VALID_POST_KEYS:
        UI.error(
            f"Post id {post.get('id')} of {post.get('user')} "
            f"has invalid schema. "
            f"Missing: {VALID_POST_KEYS - post_keys}, "
            f"Extra: {post_keys - VALID_POST_KEYS}"
        )
        return None

    try:
        domain = parser.get_domain(post["service"])
        raw_json, json_hash, fetched_at = _compute_json_metadata(post)
        return Post(
            **post,
            domain=domain,
            json_hash=json_hash,
            raw_json=raw_json,
            fetched_at=fetched_at,
        )
    except TypeError as e:
        UI.error(
            f"Post id {post.get('id')} from {post.get('user')} could not be parsed: {e}"
        )
        return None


def row_to_post(row: sqlite3.Row) -> Post | None:
    """Convert a sqlite3 row into a Post model.
    Return None if conversion failed"""
    try:
        raw = json.loads(row["raw_json"])
        return Post(
            id=row["id"],
            user=row["user"],
            service=row["service"],
            domain=row["domain"],
            published=row["published"],
            json_hash=row["json_hash"],
            raw_json=row["raw_json"],
            fetched_at=row["fetched_at"],
            title=raw["title"],
            substring=raw["substring"],
            file=raw["file"],
            attachments=raw["attachments"],
        )
    except KeyError as e:
        UI.error(
            f"KeyError: Failed to convert {row['id']} (row_id) into Post model due to: {e}"
        )
        return None
    except TypeError as e:
        UI.error(
            f"TypeError: Failed to convert {row['id']} (row_id) into Post model due to: {e}"
        )
        return None
    except ValueError as e:
        UI.error(
            f"ValueError/JSONDecodeError: Failed to convert "
            f"{row['id']} (row_id) into Post model due to: {e}"
        )
        return None


def rows_to_posts(rows: list[sqlite3.Row]) -> list[Post]:
    """Convert a list of sqlite3 rows. Return a list of Post model.
    Ignore the row if conversion fail"""
    posts: list[Post] = []
    for row in rows:
        post = row_to_post(row)
        if post is not None:
            posts.append(post)

    if len(posts) != len(rows):
        UI.error(
            f"From {len(rows)} rows, only converted {len(posts)}."
            f" {len(rows) - len(posts)} error."
        )

    return posts


def row_to_media(row: sqlite3.Row) -> Media | None:
    """Convert a sqlite3 row into a Media model.
    Return None if conversion failed"""
    try:
        # create a dict to hold column of row that are present in Media.
        # Ignore column (like default autoincrement ID) that are not a field in Media
        media_data = {}
        for field in fields(Media):
            field_name = field.name
            if field_name in row.keys():
                value = row[field_name]
                if field_name == "status" and value is not None:
                    value = Status(value)
                media_data[field_name] = value
        return Media(**media_data)
    except (KeyError, TypeError, ValueError) as e:
        UI.error(
            f"Key/Type/Value Error: Failed to convert row {row['id']} into Post model due to {e}"
        )
        return None


def rows_to_medias(rows: list[sqlite3.Row]) -> list[Media]:
    """Convert a list of sqlite3 rows. Return a list of Media model.
    Ignore row if conversion failed"""
    medias: list[Media] = []
    for row in rows:
        media = row_to_media(row)
        if media is not None:
            medias.append(media)

    if len(medias) != len(rows):
        UI.error(
            f"From {len(rows)} rows, only converted {len(medias)}."
            f" {len(rows) - len(medias)} error."
        )

    return medias


def row_to_fused_media(row: sqlite3.Row) -> FusedMedia | None:
    """Convert a sqlite3 row into a FusedMedia model.
    Return None if conversion fail"""
    if row is None:
        return None
    try:
        fuses_data = {}
        for field in fields(FusedMedia):
            field_name = field.name
            if field_name in row.keys():
                value = row[field_name]
                if field_name == "status" and value is not None:
                    value = Status(value)
                fuses_data[field_name] = value
        return FusedMedia(**fuses_data)
    except (KeyError, TypeError, ValueError) as e:
        UI.error(
            f"Key/Type/Value Error: Failed to convert row "
            f"{row['id']} into FusedMedia model due to {e}"
        )
        return None


def rows_to_fuses(rows: list[sqlite3.Row]) -> list[FusedMedia]:
    """Convert a lsit of sqlite3 rows into a list of FusedMedia model
    Ignore row if conversion failed"""
    fuses: list[FusedMedia] = []
    for row in rows:
        fuse = row_to_fused_media(row)
        if fuse is not None:
            fuses.append(fuse)

    if len(fuses) != len(rows):
        UI.error(
            f"From {len(rows)} rows, only converted {len(fuses)}."
            f" {len(rows) - len(fuses)} error."
        )

    return fuses


def post_to_videos(post: Post) -> list[Media]:
    """Extract a list of Media model from a Post model"""
    json_post = json.loads(post.raw_json)

    urls = parser.extract_video_urls(json_post)
    sequence = 0
    medias: list[Media] = []
    for url in urls:
        medias.append(
            Media(
                post_id=post.id,
                service=post.service,
                url=url,
                duration=0.0,
                sequence=sequence,
                status=Status.PENDING,
                checksum="",
                file_path=parser.get_filename(json_post, url),
                created_at="",
                updated_at="",
                file_size=0,
                fail_count=0,
            )
        )
        sequence += 1
    return medias



# core/file_io.py

"""All write/read to file function (excluding sqlite database)"""

import json


def write_json(path, data, mode="w"):
    """Write dict data to json"""
    with open(path, mode, encoding="utf-8") as f:
        json.dump(data, f, indent=4)


def load_json(path) -> dict:
    """Load data from json"""
    with open(path, "r", encoding="utf-8") as f:
        data = json.load(f)
    return data


def load_txt(path) -> list[str]:
    """Read text from a .txt file.
    Return list of stripped lines"""
    with open(path, "r", encoding="utf-8") as f:
        lines = f.readlines()
    for i, line in enumerate(lines):
        lines[i] = line.strip()
    return lines


def write_txt(path, lines: list[str] | str, mode: str = "a"):
    """Write txt to .txt file"""
    if isinstance(lines, str):
        lines = [lines]

    with open(path, mode, encoding="utf-8") as f:
        for line in lines:
            if not line.endswith("\n"):
                f.write(line + "\n")
            else:
                f.write(line)



# gui/video_server.py

from fastapi import FastAPI, Query
from fastapi.responses import FileResponse
from fastapi.middleware.cors import CORSMiddleware
import os

app = FastAPI()

# FIX: Allow the browser to load videos from a different port (CORS)
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],  # Allows all origins (safe for local/homelab use)
    allow_credentials=True,
    allow_methods=["*"],  # Allows all methods
    allow_headers=["*"],  # Allows all headers
)


@app.get("/video")
async def get_video(path: str = Query(...)):
    real_path = path

    if not os.path.isfile(real_path):
        return {"error": "File not found"}, 404

    return FileResponse(
        real_path,
        media_type="video/mp4",
        # FileResponse automatically handles Range headers,
        # but explicit Accept-Ranges is good practice.
        headers={
            "Accept-Ranges": "bytes",
        },
    )
